/*  donut-nes - v2023-12-21 - public domain (details at end of file)
    codec for texture/character data of the Nintendo Entertainment System

    By Johnathan Roatch - https://jroatch.nfshost.com/donut-nes/

    To Compile, place stb_image.h from https://github.com/nothings/stb
    the same directory as donut-nes.c and with any standard C99 compiler:
    gcc -O2 -std=c99 -DUSE_MAIN_CLI_APP -o donut-nes donut-nes.c

    or to compile without image loading features:
    gcc -O2 -std=c99 -DUSE_MAIN_CLI_APP -DDONUT_NO_STBIMAGE -o donut-nes donut-nes.c

    To use this as a single header styled library, rename this to
    "jrr-donut-nes.h", and then #define JRR_DONUT_NES_IMPLEMENTATION
    in *one* C/C++ file before the #include. For example:

    #define JRR_DONUT_NES_IMPLEMENTATION
    #include "jrr-donut-nes.h"

    This compression codec is designed for the native texture data of a
    Nintendo Entertainment System (NES) which can be briefly described as
    8x8 2bpp planer formated tiles. Each NES tile uses 16 bytes of data
    (8 bytes of a "lower bitplane" followed by 8 bytes of a "higher bitplane")
    and are often arranged together with similar characteristics such as
    sharing a particular color. A fixed decoded block size of 64 bytes
    was chosen for several reasons:
    - It fits 4 tiles NES tiles which is often put together in a "metatile"
    - It's the size of 2 rows of tilemap indexes, and the attribute table
    - less then 256 bytes, beyond which would complicate 6502 addressing modes
    - It's 8^3 bits, or 8 planes of 8x8 1bpp tiles

    The compressed block is a variable sized block with most of the key
    processing info in the first 1 or 2 bytes:
    bit 76543210
        |||||||+-- Rotate plane bits (135° reflection)
        ||||000--- All planes: 0x00
        ||||010--- Odd planes: 0x00, Even planes:  pb8
        ||||100--- Odd planes:  pb8, Even planes: 0x00
        ||||110--- All planes: pb8
        ||||001--- In another header byte, For each bit starting from MSB
        ||||         0: 0x00 plane
        ||||         1: pb8 plane
        ||||011--- In another header byte, Decode only 1 pb8 plane and
        ||||       duplicate it for each bit starting from MSB
        ||||         0: 0x00 plane
        ||||         1: duplicated plane
        ||||       If extra header byte = 0x00, no pb8 plane is decoded.
        |||+------ Even planes predict from 0xff
        ||+------- Odd planes predict from 0xff
        |+-------- Even planes = Even planes XOR Odd planes
        +--------- Odd  planes = Even planes XOR Odd planes
        00101010-- Uncompressed block of 64 bytes (bit pattern is ascii '*' )
        11111111-- End of stream
        11xxxxxx-- if last byte, ones' complement of the output bytes to trim
        01xx000x-- Reserved for future use
        10xx000x-- Reserved for future use
        xxxx1x1x-- Reserved for future use

    "0x00 planes" can be 8 bytes of 0x00 or 0xff depending on the predict bit.
    "pb8 plane" consists of a 8-bit header where each bit indicates
    duplicating the previous byte or reading a literal byte. The variant
    used here unpacks backwards.
*/

#if defined(__cplusplus)
extern "C" {
#endif

#if !defined(INCLUDE_JRR_DONUT_NES_H)
#define INCLUDE_JRR_DONUT_NES_H

#include <stddef.h>       // size_t
#include <stdint.h>       // uint8_t, ect
#include <stdbool.h>      // bool

/* Until I get a whole decoding context struct working I've essentialy
   removed and inlined donut_compress() and donut_decompress() */

/*  Reads 64 bytes from 'src' and writes 1~65 bytes to 'dst'
    'mask': is a optional bitmap the same size as src where for each bit;
    - 0 means leave the bit in src alone,
    - 1 means fill in the bit in a attempt to optimize compression.
    'cpu_limit': limits the amount of time the 6502 decoder should
      take for the block, 0 means unlimited.
    Returns: the number of bytes written to 'dst'.*/
int donut_pack_block(uint8_t* dst, const uint8_t* src, const uint8_t* mask, int cpu_limit);

/*  Reads 1~74 bytes from 'src' to then decompress
    a fixed size 64 byte block to 'dst'
	if 'cycles' not NULL, the it'll be how much cpu time the
    6502 decoder would take to decode the block
	if 'dst' is NULL, only parse the length and/or cycles
    Returns: the number of bytes read from 'src', 0 for none. */
int donut_unpack_block(uint8_t* dst, const uint8_t* src, int* cycles);

/*  reads 8 bytes from 'src' to pack 1~9 bytes into 'dst'
    'top_value' is the predicted byte before the first
    This pb8 variant is ordered from last to first byte.
    and bits 0 and 1 bits means duplicate and literal respectively
    Returns: the number of bytes written to 'dst'. */
int donut_pack_pb8(uint8_t* dst, uint64_t src, uint8_t top_value);

/*  reads 1~9 bytes from 'src' to unpack 8 bytes into 'dst'
    'top_value' is the predicted byte before the first
    Returns: the number of bytes written to 'dst'. */
int donut_unpack_pb8(uint64_t* dst, const uint8_t* src, uint8_t top_value);

#endif // INCLUDE_JRR_DONUT_NES_H

#if defined(__cplusplus)
}
#endif


#if defined(JRR_DONUT_NES_IMPLEMENTATION) || defined(USE_MAIN_CLI_APP)

#include <assert.h>

// word read/write macros from https://justine.lol/endian.html
#define READ64LE(S)                                                    \
  ((uint64_t)(255 & (S)[7]) << 070 | (uint64_t)(255 & (S)[6]) << 060 | \
   (uint64_t)(255 & (S)[5]) << 050 | (uint64_t)(255 & (S)[4]) << 040 | \
   (uint64_t)(255 & (S)[3]) << 030 | (uint64_t)(255 & (S)[2]) << 020 | \
   (uint64_t)(255 & (S)[1]) << 010 | (uint64_t)(255 & (S)[0]) << 000)

#define WRITE64LE(P, V)                        \
  ((P)[0] = (0x00000000000000FF & (V)) >> 000, \
   (P)[1] = (0x000000000000FF00 & (V)) >> 010, \
   (P)[2] = (0x0000000000FF0000 & (V)) >> 020, \
   (P)[3] = (0x00000000FF000000 & (V)) >> 030, \
   (P)[4] = (0x000000FF00000000 & (V)) >> 040, \
   (P)[5] = (0x0000FF0000000000 & (V)) >> 050, \
   (P)[6] = (0x00FF000000000000 & (V)) >> 060, \
   (P)[7] = (0xFF00000000000000 & (V)) >> 070, (P) + 8)

int donut_unpack_pb8(uint64_t* dst, const uint8_t* src, uint8_t pb8_byte)
{
	assert(dst && src);
	int y = 0;
	uint8_t pb8_flags = src[y];
	++y;
	uint64_t val = 0;
	for (int i = 0; i < 8; ++i) {
		if (pb8_flags & 0x80) {
			pb8_byte = src[y];
			++y;
		}
		pb8_flags <<= 1;
		val <<= 8;
		val |= pb8_byte;
	}
	*dst = val;
	return y;
}

int donut_pack_pb8(uint8_t* dst, uint64_t src, uint8_t pb8_byte)
{
	assert(dst);
	uint8_t pb8_flags = 0;
	int y = 1;
	for (int i = 0; i < 8; ++i) {
		uint8_t c = src >> (8*(7-i));
		if (c == pb8_byte) continue;
		dst[y] = c;
		++y;
		pb8_byte = c;
		pb8_flags |= 0x80>>i;
	}
	dst[0] = pb8_flags;
	return y;
}

static uint64_t donut_flip_plane(uint64_t plane)
{
	uint64_t result = 0;
	uint64_t t;
	int i;
	if (plane == 0xffffffffffffffff) return plane;
	if (plane == 0x0000000000000000) return plane;
	for (i = 0; i < 8; ++i) {
		t = plane >> i;
		t &= 0x0101010101010101;
		t *= 0x0102040810204080;
		t >>= 56;
		t &= 0xff;
		result |= t << (i*8);
	}
	return result;
}

// kind of a transcription of the 6502 assembly decoder.
int donut_unpack_block(uint8_t* dst, const uint8_t* src, int* cycles)
{
	assert(src);
	int y = 0;
	int x = 0;
	uint8_t block_header = src[y];
	++y;
	if (block_header == 0x00) {
		if (dst) {
			for (int i = 0; i < 64; ++i) dst[i+x] = 0x00;
		}
		if (cycles) *cycles = 1272;
		return 1;
	}
	if (block_header == 0x2a) {
		if (dst) {
			for (int i = 0; i < 64; ++i) dst[i+x] = src[i+y];
		}
		if (cycles) *cycles = 1318;
		return 65;
	}
	if (cycles) *cycles = 0;
	if (block_header >= 0xc0) return 0;  // Return to caller to let it do the block triming
	if ((block_header & 0xce) == 0x40) return 0;
	if ((block_header & 0xce) == 0x80) return 0;
	if ((block_header & 0x0a) == 0x0a) return 0; // other reserved block ranges
	if (cycles) *cycles = ((block_header & 0xc0) ? 1912 : 1272)
		+ ((block_header & 0x20) ? 4 : 0)
		+ ((block_header & 0x10) ? 4 : 0)
		+ ((block_header & 0x02) ? 5 : 0);
	uint8_t plane_def = 0xffaa5500 >> ((block_header & 0x0c) << 1);
	bool single_plane_mode = false;
	if (block_header & 0x02) {
		single_plane_mode = (block_header & 0x04);
		plane_def = src[y];
		++y;
	}
	uint64_t prev_plane = 0x0000000000000000;
	for (int i = 0; i < 8; ++i) {
		uint64_t plane = 0x0000000000000000;
		if (!(i & 1) && (block_header & 0x20)) plane = 0xffffffffffffffff;
		if ((i & 1) && (block_header & 0x10)) plane = 0xffffffffffffffff;
		if (plane_def & 0x80) {
			if (single_plane_mode) y = 2;
			int l = donut_unpack_pb8(&plane, &src[y], (uint8_t)plane);
			y += l;
			if (dst && block_header & 0x01) plane = donut_flip_plane(plane);
			if (cycles) *cycles += ((block_header & 0x01) ? 613 : 75)
				+ ((single_plane_mode) ? 1 : 0)
				+ ((l-1) * 6);
		}
		plane_def <<= 1;
		if (dst && i & 1) {
			if (block_header & 0x80) prev_plane ^= plane;
			if (block_header & 0x40) plane ^= prev_plane;
			(void)WRITE64LE(&dst[x], prev_plane);
			x += 8;
			(void)WRITE64LE(&dst[x], plane);
			x += 8;
		}
		prev_plane = plane;
	}
	return y;
}

// TODO: Clean up fill_dont_care_bits stuff?
static uint64_t donut_nes_fill_dont_care_bits_helper(uint64_t plane, uint64_t dont_care_mask, uint64_t xor_bg, uint8_t top_value) {
	uint64_t result_plane = 0;
	uint64_t backwards_smudge_plane = 0;
	uint64_t current_byte, mask, inv_mask;
	int i;
	if (dont_care_mask == 0x0000000000000000) return plane;

	current_byte = top_value;
	for (i = 0; i < 8; ++i) {
		mask = dont_care_mask & ((uint64_t)0xff << (i*8));
		inv_mask = ~dont_care_mask & ((uint64_t)0xff << (i*8));
		current_byte = (current_byte & mask) | (plane & inv_mask);
		backwards_smudge_plane |= current_byte;
		current_byte = current_byte << 8;
	}
	backwards_smudge_plane ^= xor_bg & dont_care_mask;

	current_byte = (uint64_t)top_value << 56;
	for (i = 0; i < 8; ++i) {
		mask = dont_care_mask & ((uint64_t)0xff << (8*(7-i)));
		inv_mask = ~dont_care_mask & ((uint64_t)0xff << (8*(7-i)));
		if ((plane & inv_mask) == (current_byte & inv_mask)) {
			current_byte = (current_byte & mask) | (plane & inv_mask);
		} else {
			current_byte = (backwards_smudge_plane & mask) | (plane & inv_mask);
		}
		result_plane |= current_byte;
		current_byte = current_byte >> 8;
	}

	return result_plane;
}

static void donut_nes_fill_dont_care_bits(uint64_t* planes, const uint64_t* masks, uint8_t mode)
{
	uint64_t plane_predict_l;
	uint64_t plane_predict_m;
	for (int i = 0; i < 8; i += 2) {
		plane_predict_l = (mode & 0x20) ? 0xffffffffffffffff : 0x0000000000000000;
		planes[i+0] = donut_nes_fill_dont_care_bits_helper(planes[i+0], masks[i+0], 0, plane_predict_l);
		plane_predict_m = (mode & 0x10) ? 0xffffffffffffffff : 0x0000000000000000;
		planes[i+1] = donut_nes_fill_dont_care_bits_helper(planes[i+1], masks[i+1], 0, plane_predict_m);

		if (mode & 0x80) planes[i+0] = donut_nes_fill_dont_care_bits_helper(planes[i+0], masks[i+0], planes[i+1], plane_predict_l);
		if (mode & 0x40) planes[i+1] = donut_nes_fill_dont_care_bits_helper(planes[i+1], masks[i+1], planes[i+0], plane_predict_m);
	}
	return;
}

static void donut_compare_cost_and_write_if_smaller(const uint8_t* cblock, uint8_t* dst, int* shortest_len, uint64_t* least_cost, unsigned int cpu_limit) {
	assert(cblock && dst && shortest_len && least_cost);
	// parse how much cpu time the 6502 decoder would take
	int cycles;
	int length = donut_unpack_block(NULL, cblock, &cycles);
	if (!length) return;

    // compute a hash for dithering selections of the same length and cycle cost
    uint32_t block_hash = (2166136261 ^ (uint8_t)length) * 16777619;
	for (int i = 0; i < length; ++i) block_hash = (block_hash ^ cblock[i]) * 16777619;
	block_hash = (block_hash>>16) ^ (block_hash&0xffff);

    // length (7 bits), 6502 cycles (13 bits), rotate (1 bit), hash (16 bit)
	// Adding a rotate bit after the hash makes sure that only
	// certain solid color block types are selected.
	uint64_t cost = ((uint64_t)length<<30) | (cycles<<17) | ((cblock[0] & 0x01)<<16) | block_hash;
	if ((cost < *least_cost) && (((cost & 0x3ffe0000) >> 17) <= cpu_limit)) {
		for (int i = 0; i < length; ++i) dst[i] = cblock[i];
		*shortest_len = length;
		*least_cost = cost;
	}
}

int donut_pack_block(uint8_t* dst, const uint8_t* src, const uint8_t* mask, int cpu_limit)
{
	assert(dst && src);
	uint64_t planes[8];
	uint64_t plane_masks[8];
	uint8_t cblock[74]; // a max encoded block is 2+8*9 == 74
	if (cpu_limit <= 0) cpu_limit = 8192;  // basically unlimited.

	bool all_planes_zero = true;
	for (int i = 0; i < 8; ++i) {
		plane_masks[i] = (mask) ? READ64LE(&mask[i*8]) : 0;
		planes[i] = READ64LE(&src[i*8]) & ~plane_masks[i];
		if (planes[i]) all_planes_zero = false;
	}
	// first try a zero block as it's the fastest block of all
	dst[0] = 0x00;
	if (all_planes_zero) return 1;
	// next the uncompressed block which will be the ultimate fallback
	dst[0] = 0x2a;
	for (int i = 0; i < 64; ++i) dst[i+1] = src[i];
	int shortest_len = 65;
	uint64_t least_cost = ((uint64_t)65<<30) | (1318<<17) | 0xffff;
	// if cpu_limit constrains too much, uncompressed blocks are all that can happen.
	if (cpu_limit < 1272) return shortest_len;

	// Try to compress with all 48 different block modes.
	// The rotate bit (0x01) will toggle last so that flip_plane
	// is only called once instead of 24 times.
	for (uint8_t a = 0x00; ; a += 0x10) {
		if (a >= 0xc0) {
			if (a & 0x01) break;
			for (int i = 0; i < 8; ++i) {
				planes[i] = donut_flip_plane(planes[i]);
				plane_masks[i] = donut_flip_plane(plane_masks[i]);
			}
			a = 0x01;
		}
		if (mask) donut_nes_fill_dont_care_bits(planes, plane_masks, a);
		// With the block mode in mind, pack the 64 bytes of data into 8 pb8 planes.
		uint8_t plane_def = 0x00;
		int len = 2; // make room for the block header
		int pb8_count = 0;
		uint64_t first_non_zero_plane = 0;
		int first_pb8_plane_length = 0;
		bool decoded_planes_match = true;
		for (int i = 0; i < 8; ++i) {
			uint64_t plane_predict = 0x0000000000000000;
			uint64_t plane = planes[i];
			if (i & 1) {
				if (a & 0x10) plane_predict = 0xffffffffffffffff;
				if (a & 0x40) plane ^= planes[i-1];
			} else {
				if (a & 0x20) plane_predict = 0xffffffffffffffff;
				if (a & 0x80) plane ^= planes[i+1];
			}
			plane_def <<= 1;
			if (plane != plane_predict) {
				int l = donut_pack_pb8(&cblock[len], plane, (uint8_t)plane_predict);
				len += l;
				plane_def |= 1;
				++pb8_count;
				if (pb8_count == 1) {
					first_non_zero_plane = plane;
					first_pb8_plane_length = l;
				} else if (plane != first_non_zero_plane) {
					decoded_planes_match = false;
				}
			}
		}
		if (!pb8_count) decoded_planes_match = false;

		// The only allowable block types for solid color blocks are
		// ones that do not XOR planes, those are reserved for future use
		if (!pb8_count && (a & 0xc0)) continue;

		// now try to find optimizations, the the basic block form
		// is checked in later *after* checking optimizations.
		// the order of checks is to reduce the number of memcpy

		// single plane mode reusing the already encoded block
		// We can't use decoded_planes_match here because the pb8 top byte might vary
		// also at pb8_count of 1 basic block is cheaper
		if (pb8_count > 1 && first_pb8_plane_length*pb8_count == len-2) {
			// so instead check that all encoded pb8 plane bytes match
			bool all_pb8_planes_match = true;
			for (int i = 2; i < len-first_pb8_plane_length; ++i) {
				if (cblock[i] != cblock[i+first_pb8_plane_length]) {
					all_pb8_planes_match = false;
					break;
				}
			}
			if (all_pb8_planes_match) {
				cblock[0] = a | 0x06;
				cblock[1] = plane_def;
				donut_compare_cost_and_write_if_smaller(cblock, dst, &shortest_len, &least_cost, cpu_limit);
			}
		}

		// shorthand plane_def optimization
		for (int i = 0; i < 4*8; i += 8) {
			if (plane_def != ((0xffaa5500 >> i) & 0xff)) continue;
			cblock[1] = a | (i >> 1);
			// The first byte is moved up by 1, which means the length is less 1 as well
			donut_compare_cost_and_write_if_smaller(&cblock[1], dst, &shortest_len, &least_cost, cpu_limit);
			break;
		}

		// basic unoptimized block
		cblock[0] = a | 0x02;
		cblock[1] = plane_def;
		donut_compare_cost_and_write_if_smaller(cblock, dst, &shortest_len, &least_cost, cpu_limit);

		// single plane mode where the pb8 plane leads with a 0x00/0xff
		// byte that overrides the normal top byte prediction done last
		// because pb8 planes are re-encoded. also skip if there's only
		// 1 plane as the "unoptimized block" is certainly cheaper in that case
		if (decoded_planes_match && pb8_count > 1) {
			cblock[0] = a | 0x06;
			cblock[1] = plane_def;
			len = 2 + donut_pack_pb8(&cblock[2], first_non_zero_plane, ~(first_non_zero_plane >> (7*8)));
			donut_compare_cost_and_write_if_smaller(cblock, dst, &shortest_len, &least_cost, cpu_limit);
		}
	}

	return shortest_len;
}

size_t donut_compress_bound(size_t src_size) {
	return ((((src_size) + 63) / 64) * 65);
}

#endif // JRR_DONUT_NES_IMPLEMENTATION


#if defined(USE_MAIN_CLI_APP)

// Standard headers that do not require the C runtime
#include <stddef.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <stdint.h>       // C99
#include <stdbool.h>      // C99

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(DONUT_NO_STBIMAGE)
#define STB_IMAGE_IMPLEMENTATION
#define STBI_NO_STDIO
#define STBI_NO_HDR
#define STBI_FAILURE_USERMSG
#include "stb_image.h"
#endif

const char *PROGRAM_NAME = "donut-nes";
const char *USAGE_TEXT =
	"donut-nes - A NES CHR Codec\n"
	"\n"
	"Usage:\n"
	"  [un]donut-nes [options] INPUT [OUTPUT]\n"
	"\n"
	"Options:\n"
	"  -h, --help             show this help message and exit\n"
	"  -d                     decompress input (default with \"un\" prefix)\n"
	"  -z                     compress input (default)\n"
	"  -o FILE                output to FILE instead of second positional argument\n"
	"  -f                     overwrite output without prompting\n"
#if !defined(DONUT_NO_STBIMAGE)
	"  -i                     try to load the CHR file as a image\n"
#endif
	"  -q                     suppress error messages\n"
	"  -v                     show completion stats\n"
	"  -n                     omit EOF byte in compressed output\n"
	"  -p INT                 limits the 6502 decoding time for each encoded block\n"
	"  -m                     input is byte interleaved with \"don't care bits\" mask\n"
//	"  -a                     compress input to Donut 2.0\n"
//	"  -l FILE                output a list of byte offsets for each block\n"
	"Notes:\n"
	"  use - to refer to stdin or stdout\n"
	"  If only INPUT is specified, its file extension will determine the output type\n"
;

// set_fd_binary() Sets stdin/stdout to binary mode under Windows
// Thanks to Pino
#if defined(_WIN32)
#include <io.h>
#include <fcntl.h>
#define fd_isatty _isatty
#endif
static inline void set_fd_binary(unsigned int fd) {
#if defined(_WIN32)
  _setmode(fd, _O_BINARY);
#else
  (void) fd;
#endif
}

// Reads files without fseek and ftell capabilities (stdin, named pipes)
// this does so by using realloc a bunch of times.
// returns 0 and sets data_ptr to NULL if error or zero sized file.
static size_t read_whole_file_without_fseek(uint8_t** data_ptr, FILE* f)
{
	uint8_t* data = NULL;
	size_t length = 0;
	uint8_t* realloc_data = NULL;
	// I want a slower exponential growth then something like a simple
	// "capacity *= 2", so I choose the fibonacci sequence (growth rate of about 1.6).
	// 4KiB (common memory page size) times the 6th fibonacci number is 32kiB.
	size_t capacity = 32768;
	size_t previous_capacity = 20480;

	assert(data_ptr && f);
	*data_ptr = NULL;

	data = malloc(capacity);
	if (!data) return 0;

	while (!feof(f)) {
		length += fread(data + length, sizeof(uint8_t), capacity - length, f);
		if (ferror(f)) goto free_and_return_empty;

		if (length == capacity) {
			size_t n = capacity;
			capacity = capacity + previous_capacity;
			previous_capacity = n;
			realloc_data = realloc(data, capacity);
			if (!realloc_data) goto free_and_return_empty;
			data = realloc_data;
		}
	}
	if (!length) goto free_and_return_empty;
	// shrink buffer to fit the final size
	realloc_data = realloc(data, length);
	if (!realloc_data) goto free_and_return_empty;
	data = realloc_data;

	*data_ptr = data;
	return length;
free_and_return_empty:
	free(data);
	return 0;
}

// returns 0 and sets data_ptr to NULL if error or zero sized file.
static size_t read_whole_file(uint8_t** data_ptr, FILE* f)
{
	assert(data_ptr && f);
	*data_ptr = NULL;

	if (fseek(f, 0, SEEK_END) != 0) {
		// If fseek fails the file cursor is still at the beginning
		// so no need for rewind(f)
		return read_whole_file_without_fseek(data_ptr, f);
	}

	// fseek works, so now do it the simple way
	size_t length = ftell(f);
	if (length <= 0) return 0;
	uint8_t* data = malloc(length);
	if (!data) return 0;
	rewind(f);
	size_t read_length = fread(data, sizeof(uint8_t), length, f);
	if (read_length != length) {
		free(data);
		return 0;
	}

	*data_ptr = data;
	return length;
}

int main (int argc, char* argv[])
{
	// Yes, I'm rolling my own argument parsing here
	// because some compiler didn't have <getopt.h> :(
	bool print_help_and_exit = false;
	bool decompress = false;
	bool arg_force_compress_mode = false;
	bool arg_omit_terminator = false;
	bool arg_force_overwrite = false;
	bool arg_input_mask_interleave = false;
#if !defined(DONUT_NO_STBIMAGE)
	bool arg_chr_is_image = false;
#endif
	bool arg_encode_adv_format = false;  // TBD
	int arg_number_of_filenames = 0;
	char arg_last_unknown_option = '\0';
	int arg_verbosity = 0;
	int arg_cpu_limit = 0;
	const char *input_filename = NULL;
	const char *output_filename = NULL;
	const char *stdio_filename_placeholder = "<stdio>";

	if (argc <= 1) {
		print_help_and_exit = true;
		fputs(USAGE_TEXT, stdout);
		return 0;
	}

	// check the first character of the basename of argv[0] for un-
	int arg_basename_offset = 0;
	for (int optpos = 0; argv[0][optpos] != '\0'; ++optpos) {
		if (optpos == 0 && argv[0][optpos] == '-') {
			arg_basename_offset = optpos+1;
		}
		if ((argv[0][optpos] == '/' || argv[0][optpos] == '\\') &&
			(argv[0][optpos+1] != '/' && argv[0][optpos+1] != '\\' && argv[0][optpos+1] != '\0')) {
			arg_basename_offset = optpos+1;
			arg_encode_adv_format = false;
		}
		if (argv[0][optpos] == '2') arg_encode_adv_format = true;
	}
	if (argv[0][arg_basename_offset] == 'u' || argv[0][arg_basename_offset] == 'U') decompress = true;

	bool parse_options = true;
	for (int optind = 1; optind < argc; ++optind) {
		// check to see if a option list or a old windows style single option
		bool is_option = (argv[optind][0] == '-') || 
			(argv[optind][0] == '/' && argv[optind][1] != '\0' && argv[optind][2] == '\0');
		// check for stdio shorthand "-"
		const char *filename_ptr = argv[optind];
		if (parse_options && argv[optind][0] == '-' && argv[optind][1] == '\0') {
			filename_ptr = stdio_filename_placeholder;
			is_option = false;
		}
		if (!parse_options || !is_option) {
			if (!input_filename) {
				input_filename = filename_ptr;
			} else if (!output_filename) {
				output_filename = filename_ptr;
			}
			++arg_number_of_filenames;
			continue;
		}
		for (int optpos = 1; argv[optind][optpos] != '\0'; ++optpos) {
			bool goto_next_optind = false;
			int c = argv[optind][optpos];
			if ((unsigned)c-'A' < 26) c |= 32;  // tolower()
			switch (c) {
			case '?': print_help_and_exit = true; break;
			case 'h': print_help_and_exit = true; break;
			case 'd': arg_force_compress_mode = true; decompress = true; break;
			case 'z': arg_force_compress_mode = true; decompress = false; break;
			case 'o':
				++optpos;
				if (!argv[optind][optpos] && optind+1 < argc) {
					++optind;
					optpos = 0;
				}
				output_filename = &argv[optind][optpos];
				++arg_number_of_filenames;
				goto_next_optind = true;
			break;
			case 'f': arg_force_overwrite = true; break;
#if !defined(DONUT_NO_STBIMAGE)
			case 'i': arg_chr_is_image = true; break;
#endif
			case 'q': arg_verbosity = -1; break;
			case 'v': if (arg_verbosity >= 0) ++arg_verbosity; break;
			case 'n': arg_omit_terminator = true; break;
			case 'p':
				++optpos;
				if (!argv[optind][optpos] && optind+1 < argc) {
					++optind;
					optpos = 0;
				}
				arg_cpu_limit = strtol(&argv[optind][optpos], NULL, 0);
				goto_next_optind = true;
			break;
			case 'm': arg_input_mask_interleave = true; break;
			case 'a': arg_encode_adv_format = true; break;
			case '-':
				if (optpos == 1) {
					if (argv[optind][optpos+1] == '\0') {
						// "--" ends option parsing
						parse_options = false;
					} else {
						// The only long option is "help"
						// so have all long options be just that
						print_help_and_exit = true;
						goto_next_optind = true;
					}
				} else {
					arg_last_unknown_option = c;
				}
			break;
			default: arg_last_unknown_option = c; break;
			}
			if (goto_next_optind) break;
		}
	}

	// for the single filename case, set automatic options and output filename
	char output_filename_buffer[32768];
	if (input_filename && input_filename != stdio_filename_placeholder && !output_filename) {
		// scan for the filename extention to replace
		// while keeping in mind directory separators and dot files
		int basename_begin = 0;
		int ext_begin = 0;
		int ext_end = 0;
		for (int i = 0; ; ++i) {
			int c = input_filename[i];
			if (basename_begin < i) {
				if (c == '.') ext_begin = i;
				if (c == '/' || c == '\\' || c == '\0') ext_end = i;
				if (ext_end == i && ext_begin <= basename_begin) ext_begin = i; // place extention to end of string if a dot wasn't found
				if (ext_end == i && (ext_end - ext_begin) == 1) ext_begin = i; // treat trailing dot as part of base excluded from extention
				if (ext_end == i) basename_begin = i+1; // entering diretory slash
			} else {
				if (c == '/' || c == '\\' || c == '.') basename_begin = i+1;
				// keep on punting forward the basename point to
				// inhibit sudden string end extention placement
				// this also helps collapse /./ and /../ shenanigans
			}
			if (c == '\0') break;
		}
		// the above scan can fail if filename is something like just the current directory
		if (ext_end) {
			if (!arg_force_compress_mode) decompress = false;
			int c = input_filename[ext_begin];
			if (c == '.') c = input_filename[ext_begin+1];
			if ((unsigned)c-'A' < 26) c |= 32;  // tolower()
			switch (c) {
			case 'd':
				// decode donut to raw chr or to image if arg_chr_is_image == true
				if (!arg_force_compress_mode) decompress = true;
			break;
			case 'x': // TODO: figure out the extention
				arg_input_mask_interleave = true;  // raw chr with mask
			break;
#if !defined(DONUT_NO_STBIMAGE)
			case 'b': // .bmp or .bin
				// if not .bmp then read raw chr to encode donut
				if (input_filename[ext_begin+1] != 'm') break;
			// fallthrough
			case 'j': // .jpg .jpeg .jpe .jif .jfif .jfi
			case 'p': // .png .psd .pic .pgm .ppm
			case 't': // .tga
			case 'g': // .gif
				if (!decompress) arg_chr_is_image = true;
			break;
#endif
			default:
				// all other cases read raw chr to encode donut
			break;
			}
			const char *out_file_ext = decompress ? ".chr" : ".donut";
			int k = 0;
			for (int i = 0; i < ext_begin; ++i, ++k) output_filename_buffer[k] = input_filename[i];
			for (int i = 0; out_file_ext[i]; ++i, ++k) output_filename_buffer[k] = out_file_ext[i];
			for (int i = ext_end; input_filename[i]; ++i, ++k) output_filename_buffer[k] = input_filename[i];
			output_filename_buffer[k] = '\0';

			output_filename = output_filename_buffer;
		}
	}

// print out argv parsing and argv parsing error
	if (arg_verbosity >= 3) {
		fprintf(stderr, "verbosity_level: %d\n", arg_verbosity);
		if (!input_filename) {
			fprintf(stderr, "input: null\n");
		} else if (input_filename == stdio_filename_placeholder) {
			fprintf(stderr, "input: <stdin>\n");
		} else {
			fprintf(stderr, "input: %s\n", input_filename);
		}
		fprintf(stderr, decompress ? "decompress\n" : "compress\n");
		if (!output_filename) {
			fprintf(stderr, "output: null\n");
		} else if (output_filename == stdio_filename_placeholder) {
			fprintf(stderr, "output: <stdin>\n");
		} else {
			fprintf(stderr, "output: %s\n", output_filename);
		}
		fprintf(stderr, print_help_and_exit ? "print_usage_text: true\n" : "print_usage_text: false\n");
		fprintf(stderr, "number_of_filenames: %d\n", arg_number_of_filenames);
		fprintf(stderr, "Unknown Option: '%c'\n", arg_last_unknown_option); 
		fprintf(stderr, arg_force_compress_mode ? "compress_mode_stated: true\n" : "compress_mode_stated: false\n");
		fprintf(stderr, arg_force_overwrite ? "force_file_overwrite: true\n" : "force_file_overwrite: false\n");
#if !defined(DONUT_NO_STBIMAGE)
		fprintf(stderr, arg_chr_is_image ? "input_is_image: true\n" : "input_is_image: false\n");
#else
		fprintf(stderr, "stb_image.h not compiled\n");
#endif
		fprintf(stderr, arg_input_mask_interleave ? "input_is_interleaved: true\n" : "input_is_interleaved: false\n");
		fprintf(stderr, arg_omit_terminator ? "omit_file_terminator: true\n" : "omit_file_terminator: false\n");
		fprintf(stderr, "cpu_limit: %d\n", arg_cpu_limit);
		fprintf(stderr, arg_encode_adv_format ? "donut_2: true\n" : "donut_2: false\n");
	}

	if (print_help_and_exit) {
		fputs(USAGE_TEXT, stdout);
		return 0;
	}

	if (input_filename == stdio_filename_placeholder && !output_filename) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: stdin as the input file requires a explicitly stated output file, Try --help for usage info.\n", argv[0]);
		return 1;
	}
	if (arg_number_of_filenames > 2) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: Too many filenames, Try --help for usage info.\n", argv[0]);
		return 1;
	}
	if (arg_last_unknown_option) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: Unknown option '%c', Try --help for usage info.\n", argv[0], arg_last_unknown_option);
		return 1;
	}
	if (!input_filename) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: Input filename required, Try --help for usage info.\n", argv[0]);
		return 1;
	}
	if (!output_filename) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: Could not infer required Output filename, Try --help for usage info.\n", argv[0]);
		return 1;
	}


// Read input
	FILE *input_file = NULL;
	if (input_filename != stdio_filename_placeholder) {
		input_file = fopen(input_filename, "rb");
		if (!input_file) {
			if (arg_verbosity >= 0) perror(input_filename);
			return 1;
		}
	} else {
		input_file = stdin;
		set_fd_binary(0);
	}
	uint8_t *input_data = NULL;
	uint8_t *mask_data = NULL;
	size_t input_data_size = 0;
	input_data_size = read_whole_file(&input_data, input_file);
	fclose(input_file);
// parse chr from image
#if !defined(DONUT_NO_STBIMAGE)
	if (!decompress && arg_chr_is_image) {
		int w, h;
		unsigned char *image_data = stbi_load_from_memory (input_data, input_data_size, &w, &h, NULL, 4);
		if (!image_data) {
			if (arg_verbosity >= 0) fprintf(stderr, "%s: Image load error: %s\n", argv[0], stbi_failure_reason());
			return 1;
		}
		free(input_data);
		input_data_size = ((w/8)*(h/8))*16;
		input_data = calloc(input_data_size, sizeof(uint8_t));
		mask_data = calloc(input_data_size, sizeof(uint8_t));
		if (!input_data || !mask_data) {
			if (arg_verbosity >= 0) fprintf(stderr, "%s: Out of memory\n", argv[0]);
			return 1;
		}
		for (int ty = 0; ty < h/8; ++ty) {
		for (int py = 0; py < 8; ++py) {
		for (int tx = 0; tx < w/8; ++tx) {
		for (int px = 0; px < 8; ++px) {
			size_t pixel_offset = ((ty*8)+py)*w + ((tx*8)+px);
			size_t tile_offset = (ty*(w/8)+tx)*16;
			int r = image_data[pixel_offset*4+0];
			int g = image_data[pixel_offset*4+1];
			int b = image_data[pixel_offset*4+2];
			// using BT.709/sRGB Luma coefficients of (0.2126, 0.7152, 0.0722)
			// but without gamma so that I can avoid linking math.h for pow()
			int bits = (int)((float)r * 0.003322 + (float)g * 0.011175 + (float)b * 0.001128);
			// center the rgb point to check so that the center of
			// spheres are between pixels, and divisible by 3
			r = (r*2+1)*3;
			g = (g*2+1)*3;
			b = (b*2+1)*3;
			const int pix_tbl[6][4] = { // {r*6, g*6, b*6, bits}
				{1536,    0, 1536, 0xc}, // hot pink for don't care
				{   0, 1024, 1536, 0x8}, // special lower plane only pixels
				{1536, 1024,    0, 0x9},
				{1536,    0,    0, 0x1},  // NES Maker RGB style
				{   0, 1536,    0, 0x2}, 
				{   0,    0, 1536, 0x3},
			};
			for (int i = 0; i < 6; ++i) {
				if ((r-pix_tbl[i][0])*(r-pix_tbl[i][0]) + (g-pix_tbl[i][1])*(g-pix_tbl[i][1]) + (b-pix_tbl[i][2])*(b-pix_tbl[i][2]) < 512*512) bits = pix_tbl[i][3];
			}
			if (image_data[pixel_offset*4+3] < 0x80) bits |= 0xc;

			if (bits & 1) input_data[tile_offset + py + 0] |= 0x80 >> px;
			if (bits & 2) input_data[tile_offset + py + 8] |= 0x80 >> px;
			if (bits & 4) mask_data[tile_offset + py + 0] |= 0x80 >> px;
			if (bits & 8) mask_data[tile_offset + py + 8] |= 0x80 >> px;
		}}}}
		// Trim down file size according to trailing don't care bits
		while (input_data_size > 0 && mask_data[input_data_size-1] == 0xff) --input_data_size;
		stbi_image_free(image_data);
		arg_input_mask_interleave = false; // We parsed the pixels to this form already
	}
#endif
	// if the input is a interleaved file, unzip it into 2 buffers
	if (!decompress && arg_input_mask_interleave) {
		size_t new_data_size = (input_data_size+1)/2;
		mask_data = malloc(new_data_size);
		if (!mask_data) {
			if (arg_verbosity >= 0) fprintf(stderr, "%s: Out of memory\n", argv[0]);
			return 1;
		}
		for (size_t i = 0; i*2 < input_data_size; ++i) {
			input_data[i] = input_data[i*2+0];
			if (i*2+1 < input_data_size) mask_data[i] = input_data[i*2+1];
		}
		input_data_size = new_data_size;
		// TODO: round up the buffers to the next 64 bytes
	}

	(void)arg_encode_adv_format; // TBD

// TODO: the buffering strategy is basically like read_whole_file_without_fseek
// will try to change and move back out to public functions some day later
	uint8_t* output_data = NULL;
	size_t output_data_size = 0;
	size_t output_data_capacity = 32768;
	size_t previous_capacity = 20480;

	output_data = malloc(output_data_capacity);
	if (!output_data) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: Out of memory\n", argv[0]);
		return 1;
	}

	// add bytes to buffer to allow decoding past the end,
	// so that we don't have to copy to temp buffers to give that warning.
	// TODO: make the extra bytes so that for encoding it's multipules of 64 bytes
	//       and for decoding it's a additional 74 bytes 
	uint8_t* realloc_data = realloc(input_data, input_data_size+128);
	if (!realloc_data) {
		if (arg_verbosity >= 0) fprintf(stderr, "%s: Out of memory\n", argv[0]);
		return 1;
	}
	input_data = realloc_data;
	for (size_t i = input_data_size; i < input_data_size+128; ++i) input_data[i] = 0xff;

	if (mask_data) {
		uint8_t* realloc_data = realloc(mask_data, input_data_size+128);
		if (!realloc_data) {
			if (arg_verbosity >= 0) fprintf(stderr, "%s: Out of memory\n", argv[0]);
			return 1;
		}
		mask_data = realloc_data;
		for (size_t i = input_data_size; i < input_data_size+128; ++i) mask_data[i] = 0xff;
	}

// do the encode/decode
	size_t bytes_read = 0;
	while (1) {
		if (output_data_capacity-128 < output_data_size) {
			size_t n = output_data_capacity;
			output_data_capacity = output_data_capacity + previous_capacity;
			previous_capacity = n;
			uint8_t* realloc_data = realloc(output_data, output_data_capacity);
			if (!realloc_data) {
				if (arg_verbosity >= 0) fprintf(stderr, "%s: Out of memory\n", argv[0]);
				return 1;
			}
			output_data = realloc_data;
		}

		if (decompress) {
			// reading past the end of input ensured by expanding the buffer above
			int l = donut_unpack_block(&output_data[output_data_size], &input_data[bytes_read], NULL);
			if (!l) {
				// TODO: give warning that we decoded past the end of file
				// and also make sure that the end of file is not
				// counted as part of the file size
				output_data_size -= input_data[bytes_read]^0xff;
				bytes_read += 1;
				break;
			}
			if (arg_verbosity >= 2) fprintf(stderr, "%zd\n", bytes_read);
			bytes_read += l;
			output_data_size += 64;
		} else {
			// reading past the end of input ensured by expanding the buffer above
			int l;
			if (mask_data) {
				l = donut_pack_block(&output_data[output_data_size], &input_data[bytes_read], &mask_data[bytes_read], arg_cpu_limit);
			} else if (input_data_size <= bytes_read+64) {
				uint8_t temp_mask[64] = {0};
				int bytes_to_remove = (bytes_read+64)-input_data_size;
				for (int i = 64-bytes_to_remove; i < 64; ++i) temp_mask[i] = 0xff;
				l = donut_pack_block(&output_data[output_data_size], &input_data[bytes_read], temp_mask, arg_cpu_limit);
			} else {
				l = donut_pack_block(&output_data[output_data_size], &input_data[bytes_read], NULL, arg_cpu_limit);
			}
			if (arg_verbosity >= 2) fprintf(stderr, "%zd\n", output_data_size);
			bytes_read += 64;
			output_data_size += l;
			if (input_data_size <= bytes_read) {
				int bytes_to_remove = bytes_read-input_data_size;
				if (!arg_omit_terminator) {
					output_data[output_data_size] = '\xff'-bytes_to_remove;
					output_data_size += 1;
				}
				bytes_read -= bytes_to_remove;
				break;
			}
		}
	}

// Write output data
	FILE *output_file = NULL;
	if (output_filename != stdio_filename_placeholder) {
		if (!arg_force_overwrite) {
			// open output for read to check for file existence.
			// if so do interactive question about overwriting the file
			output_file = fopen(output_filename, "rb");
			if (output_file && arg_verbosity >= 0) {
				if (input_filename == stdio_filename_placeholder) {
					fprintf(stderr, "%s already exists; not overwriting\n", output_filename);
					return 1;
				}
				fprintf(stderr, "%s already exists; do you wish to overwrite (y/N) ? ", output_filename);
				int c = fgetc(stdin);
				while (fgetc(stdin) != '\n') {;} // read until the newline
				if (c == 'y' || c == 'Y') {
					// close file reading to make avaliable for writing.
					fclose(output_file);
					output_file = NULL;
				} else {
					fprintf(stderr, "    not overwritten\n");
				}
			}
			// output file being open for reading means it's not useable for writing.
			if (output_file) return 1;
		}
		output_file = fopen(output_filename, "wb");
		if (!output_file) {
			if (arg_verbosity >= 0) perror(output_filename);
			return 1;
		}
	} else {
		output_file = stdout;
		set_fd_binary(1);
	}
	fwrite(output_data, sizeof(uint8_t), output_data_size, output_file);
	if (ferror(output_file)) {
		if (arg_verbosity >= 0) perror(output_filename);
		return 1;
	}
	fclose(output_file);

	if (arg_verbosity >= 1) {
		float total_bytes_ratio = 0.0;
		if (decompress) {
			if (output_data_size != 0) {
				total_bytes_ratio = (1.0 - ((float)bytes_read / (float)output_data_size))*100.0;
			}
		} else {
			if (output_data_size != 0) {
				total_bytes_ratio = (1.0 - ((float)output_data_size / (float)bytes_read))*100.0;
			}
		}
		fprintf (stderr, "%s :%#5.1f%% (%zd => %zd bytes)\n", output_filename, total_bytes_ratio, bytes_read, output_data_size);
	}

	free(input_data);
	free(output_data);
	free(mask_data);
	return 0;
}

#endif // USE_MAIN_CLI_APP

/*
This is free and unencumbered software released into the public domain.

Anyone is free to copy, modify, publish, use, compile, sell, or
distribute this software, either in source code form or as a compiled
binary, for any purpose, commercial or non-commercial, and by any
means.

In jurisdictions that recognize copyright laws, the author or authors
of this software dedicate any and all copyright interest in the
software to the public domain. We make this dedication for the benefit
of the public at large and to the detriment of our heirs and
successors. We intend this dedication to be an overt act of
relinquishment in perpetuity of all present and future rights to this
software under copyright law.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

For more information, please refer to <http://unlicense.org>
*/
